import time

from nose.tools import *

import voltron
from voltron.core import *
from voltron.api import *
from voltron.plugin import *

log = logging.getLogger('tests')


class APITestRequest(APIRequest):
    _fields = {'target_id': False, 'address': False, 'count': True}
    _types = {'target_id': int, 'address': int, 'count': int}

    target_id = 0
    address = None
    count = None


class APITestResponse(APISuccessResponse):
    _fields = {'disassembly': True}


class APITestPlugin(APIPlugin):
    request = 'test'
    request_class = APITestRequest
    response_class = APITestResponse


def setup():
    voltron.setup_env()


def teardown():
    time.sleep(2)


def test_parent_message_validation_fail():
    msg = APIMessage()
    exception = False
    try:
        msg.validate()
    except MissingFieldError:
        exception = True
    assert exception


def test_parent_request_validation_fail():
    msg = APIRequest()
    exception = False
    try:
        msg.validate()
    except MissingFieldError:
        exception = True
    assert exception


def test_parent_request_type():
    msg = APIRequest()
    assert msg.type == 'request'


def test_parent_request_request():
    msg = APIRequest()
    assert msg.request is None


def test_parent_response_validation_fail():
    msg = APIResponse()
    exception = False
    try:
        msg.validate()
    except MissingFieldError:
        exception = True
    assert exception


def test_parent_response_type():
    msg = APIResponse()
    assert msg.type == 'response'


def test_parent_response_status():
    msg = APIResponse()
    assert msg.status is None


def test_success_response_validation_succeed():
    msg = APISuccessResponse()
    exception = False
    try:
        msg.validate()
    except MissingFieldError:
        exception = True
    assert not exception


def test_success_response_type():
    msg = APISuccessResponse()
    assert msg.type == 'response'


def test_success_response_status():
    msg = APISuccessResponse()
    assert msg.status == 'success'


def test_error_response_validation_fail():
    msg = APIErrorResponse()
    exception = False
    try:
        msg.validate()
    except MissingFieldError:
        exception = True
    assert exception


def test_error_response_type():
    msg = APIErrorResponse()
    assert msg.type == 'response'


def test_error_response_status():
    msg = APIErrorResponse()
    assert msg.status == 'error'


def test_invalid_request_error_response_validation_succeed():
    msg = APIInvalidRequestErrorResponse()
    exception = False
    try:
        msg.validate()
    except MissingFieldError:
        exception = True
    assert not exception


def test_invalid_request_error_response_type():
    msg = APIInvalidRequestErrorResponse()
    assert msg.type == 'response'


def test_invalid_request_error_response_status():
    msg = APIInvalidRequestErrorResponse()
    assert msg.status == 'error'


def test_test_request_validation_fail():
    msg = APITestRequest()
    exception = False
    try:
        msg.validate()
    except MissingFieldError:
        exception = True
    assert exception


def test_test_request_validation_fail_with_param():
    msg = APITestRequest(target_id=0)
    exception = False
    try:
        msg.validate()
    except MissingFieldError:
        exception = True
    assert exception


def test_test_request_validation_succeed_with_param():
    msg = api_request('test', count=16)
    exception = False
    try:
        msg.validate()
    except MissingFieldError:
        exception = True
    assert not exception
    assert msg.count == 16


def test_test_request_validation_succeed_with_data():
    msg = APITestRequest('{"data":{"count":16}}')
    exception = False
    try:
        msg.validate()
    except MissingFieldError as e:
        exception = True
    assert not exception
    assert msg.count == 16


def test_test_request_validation_succeed_by_assign():
    msg = APITestRequest()
    msg.count = 16
    exception = False
    try:
        msg.validate()
    except MissingFieldError as e:
        exception = True
    assert not exception
    assert msg.count == 16


def test_test_request_string():
    msg = APITestRequest(count=16)
    assert json.loads(str(msg)) == {"request": "test", "type": "request", "block": False, "timeout": 10,
                                    "data": {"count": 16, "target_id": 0, "address": None}}


def test_test_response_validation_fail():
    msg = APITestResponse()
    exception = False
    try:
        msg.validate()
    except MissingFieldError as e:
        exception = True
    assert exception


def test_test_response_validation_fail_with_param():
    msg = APITestResponse(thing=1)
    exception = False
    try:
        msg.validate()
    except MissingFieldError as e:
        exception = True
    assert exception


def test_test_response_validation_succeed_with_param():
    msg = APITestResponse(disassembly="xxx")
    exception = False
    try:
        msg.validate()
    except MissingFieldError as e:
        exception = True
    assert not exception


def test_test_response_validation_succeed_with_data():
    msg = APITestResponse('{"data":{"disassembly":"xxx"}}')
    exception = False
    try:
        msg.validate()
    except MissingFieldError as e:
        exception = True
    assert not exception


def test_test_response_validation_succeed_by_assign():
    msg = APITestResponse()
    msg.disassembly = "xxx"
    exception = False
    try:
        msg.validate()
    except MissingFieldError as e:
        print(str(e))
        exception = True
    assert not exception


def test_test_response_string():
    msg = APITestResponse(disassembly='xxx')
    assert json.loads(str(msg)) == {"status": "success", "type": "response", "data": {"disassembly": "xxx"}}


class APIEncodeMsg(APIMessage):
    _fields = {'enc': False}
    _encode_fields = ['enc']


def test_encode_fields():
    msg = APIEncodeMsg()
    msg.enc = six.b('').join([six.int2byte(x) for x in range(0x0, 0xff)])
    assert msg.to_dict()['data']['enc'] == six.text_type('AAECAwQFBgcICQoLDA0ODxAREhMUFRYXGBkaGxwdHh8gISIjJCUmJygpKissLS4vMDEyMzQ1Njc4OTo7PD0+P0BBQkNERUZHSElKS0xNTk9QUVJTVFVWV1hZWltcXV5fYGFiY2RlZmdoaWprbG1ub3BxcnN0dXZ3eHl6e3x9fn+AgYKDhIWGh4iJiouMjY6PkJGSk5SVlpeYmZqbnJ2en6ChoqOkpaanqKmqq6ytrq+wsbKztLW2t7i5uru8vb6/wMHCw8TFxsfIycrLzM3Oz9DR0tPU1dbX2Nna29zd3t/g4eLj5OXm5+jp6uvs7e7v8PHy8/T19vf4+fr7/P3+')
    assert msg.to_json() == '{"data": {"enc": "AAECAwQFBgcICQoLDA0ODxAREhMUFRYXGBkaGxwdHh8gISIjJCUmJygpKissLS4vMDEyMzQ1Njc4OTo7PD0+P0BBQkNERUZHSElKS0xNTk9QUVJTVFVWV1hZWltcXV5fYGFiY2RlZmdoaWprbG1ub3BxcnN0dXZ3eHl6e3x9fn+AgYKDhIWGh4iJiouMjY6PkJGSk5SVlpeYmZqbnJ2en6ChoqOkpaanqKmqq6ytrq+wsbKztLW2t7i5uru8vb6/wMHCw8TFxsfIycrLzM3Oz9DR0tPU1dbX2Nna29zd3t/g4eLj5OXm5+jp6uvs7e7v8PHy8/T19vf4+fr7/P3+"}, "type": null}'

    msg2 = APIEncodeMsg(data=msg.to_json())
    assert msg.to_dict() == msg2.to_dict()
    assert msg.to_json() == msg2.to_json()
    assert msg2.enc == msg.enc

    msg3 = APIEncodeMsg()
    msg3.enc = six.u('xxxx')
    assert msg3.to_dict() == {'data': {'enc': 'eHh4eA=='}, 'type': None}
    msg3.enc = six.b('xxxx')
    assert msg3.to_dict() == {'data': {'enc': 'eHh4eA=='}, 'type': None}

    msg4 = APIEncodeMsg()
    msg4.from_dict(msg.to_dict())
    assert msg4.to_dict() == msg.to_dict()



import os
import logging
import logging.config

import voltron
from .main import main

from scruffy import Environment, Directory, File, ConfigFile, PluginDirectory, PackageDirectory

# scruffy environment containing config, plugins, etc
env = None
config = None

debugger = None
command = None
commands = None
server = None

loaded = False


def setup_env():
    global env, config
    env = Environment(setup_logging=False,
        voltron_dir=Directory('~/.voltron', create=True,
            config=ConfigFile('config', defaults=File('config/default.cfg', parent=PackageDirectory()), apply_env=True),
            sock=File('{config:server.listen.domain}'),
            history=File('history'),
            user_plugins=PluginDirectory('plugins')
        ),
        pkg_plugins=PluginDirectory('plugins', parent=PackageDirectory())
    )
    config = env.config

    voltron.plugin.pm.register_plugins()

LOGGER_DEFAULT = {
    'handlers': ['null'],
    'level': 'DEBUG',
    'propagate': False
}

LOG_CONFIG = {
    'version': 1,
    'formatters': {
        'standard': {'format': 'voltron: [%(levelname)s] %(message)s'}
    },
    'handlers': {
        'default': {
            'class': 'logging.StreamHandler',
            'formatter': 'standard'
        },
        'null': {
            'class': 'logging.NullHandler'
        }
    },
    'loggers': {
        '':         LOGGER_DEFAULT,
        'debugger': LOGGER_DEFAULT,
        'core':     LOGGER_DEFAULT,
        'main':     LOGGER_DEFAULT,
        'api':      LOGGER_DEFAULT,
        'view':     LOGGER_DEFAULT,
        'plugin':   LOGGER_DEFAULT,
    }
}


def setup_logging(logname=None):
    # configure logging
    logging.config.dictConfig(LOG_CONFIG)

    # enable the debug_file in all the loggers if the config says to
    if config and 'general' in config and config['general']['debug_logging']:
        if logname:
            filename = '{}.log'.format(logname)
        else:
            filename = 'voltron.log'
        for name in LOG_CONFIG['loggers']:
            h = logging.FileHandler(voltron.env.voltron_dir.path_to(filename), delay=True)
            h.setFormatter(logging.Formatter(fmt="%(asctime)s %(levelname)-7s %(filename)12s:%(lineno)-4s %(funcName)20s -- %(message)s"))
            logging.getLogger(name).addHandler(h)

    logging.info("======= VOLTRON - DEFENDER OF THE UNIVERSE [debug log] =======")

    return logging.getLogger(logname)


# Python 3 shim
if not hasattr(__builtins__, "xrange"):
    xrange = range


# Setup the Voltron environment
setup_env()



import sys
from setuptools import setup, find_packages


requirements = [
    'scruffington>=0.3.2',
    'flask',
    'flask_restful',
    'blessed',
    'pygments',
    'requests',
    'requests_unixsocket',
    'six'
]
if sys.platform == 'win32':
    requirements.append('cursor')

setup(
    name="voltron",
    version="0.1.4",
    author="snare",
    author_email="snare@ho.ax",
    description=("A debugger UI"),
    license="MIT",
    keywords="voltron debugger ui gdb lldb vdb vivisect vtrace windbg cdb pykd",
    url="https://github.com/snare/voltron",
    packages=find_packages(exclude=['tests', 'examples']),
    install_requires=requirements,
    package_data={'voltron': ['config/*']},
    entry_points={
        'console_scripts': ['voltron=voltron:main']
    },
    zip_safe=False
)



from voltron.view import *
from voltron.plugin import *
from voltron.api import *
try:
    from voltron.lexers import *
    have_pygments = True
except ImportError:
    have_pygments = False


class DisasmView(TerminalView):
    @classmethod
    def configure_subparser(cls, subparsers):
        sp = subparsers.add_parser('disasm', help='disassembly view', aliases=('d', 'dis', 'disasm'))
        VoltronView.add_generic_arguments(sp)
        sp.set_defaults(func=DisasmView)
        sp.add_argument('--use-capstone', '-c',  action='store_true', default=False, help='use capstone')

    def render(self):
        height, width = self.window_size()

        # Set up header & error message if applicable
        self.title = '[disassembly]'

        # Request data
        req = api_request('disassemble', block=self.block, use_capstone=self.args.use_capstone)
        req.count = self.body_height()
        res = self.client.send_request(req)

        # don't render if it timed out, probably haven't stepped the debugger again
        if res.timed_out:
            return

        if res and res.is_success:
            # Get the disasm
            disasm = res.disassembly
            disasm = '\n'.join(disasm.split('\n')[:self.body_height()])

            # Pygmentize output
            if have_pygments:
                try:
                    host = 'capstone' if self.args.use_capstone else res.host
                    lexer = all_lexers['{}_{}'.format(host, res.flavor)]()
                    disasm = pygments.highlight(disasm, lexer, pygments.formatters.TerminalFormatter())
                except Exception as e:
                    log.warning('Failed to highlight disasm: ' + str(e))

            # Build output
            self.body = disasm.rstrip()
        else:
            log.error("Error disassembling: {}".format(res.message))
            self.body = self.colour(res.message, 'red')

        # Call parent's render method
        super(DisasmView, self).render()


class DisasmViewPlugin(ViewPlugin):
    plugin_type = 'view'
    name = 'disassembly'
    aliases = ('d', 'dis', 'disasm')
    view_class = DisasmView
